{$X+}
{$M 16384, 65536, 655360}

uses
  Dos, OpKey, OpString, OpInline, OpRoot, OpDos, OpSort, OpCrt;

const
  BufferSize = 63488;

type
  FileRec = record
    Name : String [12];
    Size : Word;
  end;

type
  FileNodePtr = ^FileNode;
  FileNode = object(SingleListNode)
    Name : String;
    Size : Word;
    { ... }
    constructor Init(N : String; S : LongInt);
  end;

var
  SourcePath        : String;
  Best, Current     : BitSet;
  BestSize          : Word;
  CurrentSize       : Word;
  FileList          : SingleList;
  TargetDrive       : Char;
  ClusterSize       : Word; {     }
  DriveSize         : Word; { P ᪠   }

constructor FileNode.Init(N : String; S : LongInt);
  begin
    SingleListNode.Init;
    Name := N;
    Size := S;
  end; { FileNode.Init }

function Str2Chr(S : String) : Char;
  begin
    Str2Chr := UpCase(S[1]);
  end; { Str2Chr }

procedure GetDriveInfo(Drive : Char);
  var
    TotalClusters     : Word;
    BytesPerSector    : Word;
    SectorsPerCluster : Word;
  begin
{$IFNDEF Debug}
    GetDiskInfo(Byte(TargetDrive)-Byte('A')+1, DriveSize, TotalClusters, BytesPerSector, SectorsPerCluster);
    ClusterSize := BytesPerSector * SectorsPerCluster;
{$ELSE}
    DriveSize := 2847;
    ClusterSize := 512;
{$ENDIF}
  end; { GetDriveInfo }

procedure SendFiles; far;
  var
    SR : SearchRec;
    FR : FileRec;
  begin
    FindFirst (ParamStr(1), Archive, SR);
    while DosError = 0 do begin
      FR.Name := SR.Name;
      FR.Size := SR.Size div ClusterSize;
      if (SR.Size mod ClusterSize) > 0 then Inc(FR.Size);
      PutElement(FR);
      FindNext(SR);
    end; { while }
  end; { SendFiles }

procedure ReceiveFiles; far;
  var
    i  : Word;
    FR : FileRec;
  begin
    FileList.Init;
    for i := 1 to ElementsSorted do begin
      GetElement(FR);
      FileList.Append (New(FileNodePtr, Init(FR.Name, FR.Size)));
    end; { for }
  end; { ReceiveFiles }

function LessFile(var X, Y) : Boolean; far;
  begin
    LessFile := FileRec(X).Size > FileRec(Y).Size;
  end; { LessFile }

function YesNo(Message : String; Default : Char) : Boolean;
  var
    Ch  : Char;
  begin
    Write(Message, ' ');
    repeat
      Ch := Upcase(Char(ReadKeyWord));
      if Byte(Ch) = Lo(Enter) then Ch := Default;
    until (Ch = 'Y') or (Ch = 'N');
    WriteLn(Ch);
    YesNo := Ch = 'Y';
  end; { YesNo }

procedure Copy(var S : BitSet; FileSize : LongInt);
  var
    i : Word;
    P : Pointer;
    f : file;
    FN : FileNodePtr;
    Size : Word;
    Status : Integer;
    SaveVerify : Boolean;
  begin
    for i := 1 to FileList.Size do begin
      if S.BitIsSet(i) then begin
        FN := FileNodePtr(FileList.Find(i));
        WriteLn(Pad(FN^.Name, 20), ' - ', FN^.Size:6);
      end; { if }
    end; { for }
    WriteLn('*** 騩 ࠧ: ', FileSize, ' ஢ (', (FileSize*100) div DriveSize, '%)');
    if not YesNo('஢? [Y/n]','Y') then Exit;
    Size := MinLong (BufferSize, MaxAvail);
    Size := (Size div ClusterSize) * ClusterSize;
    if (Size = 0) or not GetMemCheck(P, Size) then begin
      WriteLn('筮 ');
      Halt;
    end; { if }
    GetVerify(SaveVerify);
    SetVerify(True);
    for i := 1 to FileList.Size do begin
      if S.BitIsSet(i) then begin
        FN := FileNodePtr(FileList.Find(i));
        Write(Pad(FN^.Name, 20), ' - ');
        Status := CopyFile(SourcePath+FN^.Name, TargetDrive+':\'+FN^.Name, P, BufferSize);
        case Status of
          0 :  WriteLn('Ok.')
          else WriteLn('FAIL (', Status, ')');
        end; { case }
      end; { if }
    end; { for }
    SetVerify(SaveVerify);
    FreeMemCheck(P, Size);
    if not YesNo(' 䠩? [y/N]', 'N') then Exit;
    for i := 1 to FileList.Size do begin
      if S.BitIsSet(i) then begin
        FN := FileNodePtr(FileList.Find(i));
        Write(Pad(FN^.Name, 20), ' - ');
        Assign(f, SourcePath+FN^.Name);
        Erase(f);
        if IOResult = 0 then
          WriteLn('Ok.')
        else
          WriteLn('FAIL!');
      end; { if }
    end; { for }
  end; { Copy }

procedure Append(N : Word);
  const
    Level : Word = 0;
  var
    FN : FileNodePtr;
  begin
    Inc(Level);
    if Level = 1 then Write('.');
    FN := FileNodePtr(FileList.Find(N));
    while (FN^.Size + CurrentSize > DriveSize) and (N <= FileList.Size) do begin
      Inc(N);
      FN := FileNodePtr(FileList.Next(FN));
    end; { while }
    if N > FileList.Size then begin
      if BestSize < CurrentSize then begin
        BestSize := CurrentSize;
        Best.Done;
        Best.Copy(Current);
      end; { if }
      Dec(Level);
      Exit;
    end; { if }
    while (N <= FileList.Size) and (BestSize < DriveSize) do begin
      Inc(CurrentSize, FN^.Size);
      Current.SetBit(N);
      Append(N+1);
      Current.ClearBit(N);
      Dec(CurrentSize, FN^.Size);
      FN := FileNodePtr(FileList.Next(FN));
      Inc(N);
    end; { while }
    Dec(Level);
  end; { Append }

begin
  WriteLn('FILLER 1.01. ()  .., 1993-95.');
  if ParamCount < 2 then begin
    WriteLn ('Usage: FILLER <source_path> <target_drive>');
    Halt;
  end; { if }
  SourcePath := AddBackSlash(JustPathName(ParamStr(1)));
  TargetDrive := Str2Chr(ParamStr(2));
  Write('뢠 ଠ  ᪥ ', TargetDrive, ': ...'^M);
  GetDriveInfo(TargetDrive);
  WriteLn(Pad('P : '+Long2Str(ClusterSize)+' .', 79));
  WriteLn(' ', DriveSize, ' ஢ (', LongInt(DriveSize)*ClusterSize, ' ).');
  Sort(1000, SizeOf(FileRec), SendFiles, LessFile, ReceiveFiles);
  Best.Init(FileList.Size);
  Current.Init(FileList.Size);
  BestSize := 0;
  CurrentSize := 0;
  Append(1);
  WriteLn;
  Copy(Best, BestSize);
end.
